from django.conf import settings
from drf_yasg.views import get_schema_view
from drf_yasg import openapi
from rest_framework import permissions


schema_view = get_schema_view(
   openapi.Info(
      title=getattr(settings, 'WD_API_TITLE', 'Wikidata API'),
      default_version=getattr(settings, 'WD_API_DEFAULT_VERSION', 'v0'),
      description=getattr(settings, 'WD_API_DESCRIPTION', '"REST API Layer Connecting To Wikidata.org'),
      terms_of_service=getattr(settings, 'WD_API_TERMS'),
      contact=openapi.Contact(name=getattr(settings, 'WD_API_CONTACT_NAME'),
                              email=getattr(settings, 'WD_API_CONTACT_EMAIL'),
                              url=getattr(settings, 'WD_API_CONTACT_URL')),
      license=openapi.License(name=getattr(settings, 'WD_API_LICENSE_NAME')),
   ),
   public=True,
   permission_classes=(permissions.AllowAny, permissions.IsAuthenticatedOrReadOnly),
)



from django.conf.urls import (
    include,
    url
)
from drf_yasg import openapi
from drf_yasg.utils import swagger_auto_schema
from rest_framework import (
    routers,
    viewsets
)
from rest_framework.decorators import action
from rest_framework.exceptions import NotFound
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response


def generate_wikidata_item_viewset(wikidata_model, slug='wikidata_item'):
    name = getattr(wikidata_model, 'model_name', "Wikidata Item")
    # TODO: Add Pluralize function
    name_plural = getattr(wikidata_model, 'model_name_plural', "{}s".format(name))
    serializer = wikidata_model.build_serializer()

    class WikidataItemViewSet(viewsets.ViewSet):
        """
        A ViewSet for listing and retrieving Wikidata Item objects.
        """
        serializer_class = serializer
        model = wikidata_model
        lookup_field = 'wikidata_id'
        permission_classes = [IsAuthenticated, ]

        @swagger_auto_schema(operation_summary="List All {}".format(name_plural),
                             operation_description="Get all {} from querying Wikidata using the Wikidata "
                                                   "SPARQL endpoint.".format(name_plural),
                             responses={200: serializer(many=True)},
                             tags=[name_plural])
        def list(self, request):
            """
            Get all objects.
            Args:
                request:

            Returns:

            """
            with_conformance = request.query_params.get('conformance', False)
            queryset = self.model.get_all(with_conformance=with_conformance)
            _serializer = self.serializer_class(queryset, many=True)
            return Response(_serializer.data)

        @swagger_auto_schema(operation_summary="Get {} By QID".format(name),
                             operation_description="Retrieve a {} by querying Wikidata for the corresponding Item's "
                                                   "Statements.".format(name),
                             responses={200: serializer(), 404: "No File Format Found with ID: Format Id"},
                             tags=[name_plural])
        def retrieve(self, request, wikidata_id=None):
            """
            Return File Format metadata if the provided QID is found in Wikidata and the item's statements matches the
            File Format schema.

            **Note:** *For more details on how we determine the File Format Schema, [see here][ref].*

            [ref]: http://example.com/activating-accounts
            """
            with_conformance = request.query_params.get('conformance', False)
            obj = self.model.get(wikidata_id, with_conformance=with_conformance)
            if not obj:
                raise NotFound(detail="No File Format Found with ID: {}".format(wikidata_id))
            _serializer = self.serializer_class(obj)
            return Response(_serializer.data)

        @swagger_auto_schema(operation_summary="Search for {}".format(name_plural),
                             operation_description="Perform a keyword lookup of metadata from all {} using the "
                                                   "Wikidata SPARQL endpoint".format(name_plural),
                             manual_parameters=[
                                 openapi.Parameter(name="query", in_="query", required=True, type="string",
                                                   description='Search For {}'.format(name_plural)),
                             ],
                             responses={200: serializer(many=True), 404: "No Search Results Found for 'query'"},
                             tags=[name_plural])
        @action(detail=False, methods=['POST'])
        def search(self, request):
            """
            Args:
                request:

            Returns:

            """
            # TODO: Add 404
            search_string = request.query_params.get('query')
            queryset = self.model.search(search_string)
            if not queryset:
                raise NotFound(detail="No Search Results Found for '{}'".format(search_string))
            _serializer = self.serializer_class(queryset, many=True)
            return Response(_serializer.data)

        @classmethod
        def get_viewset_urls(cls):
            # TODO: Possibly add slugify function to pull from class name
            router = routers.DefaultRouter()
            router.register(r'{}'.format(slug), cls, base_name=slug)
            return url(r'^'.format(slug), include(router.urls), name=slug)

    return WikidataItemViewSet.get_viewset_urls()



# coding=utf-8
""" Django-Style Wikidata Models. """
import logging
from rest_framework.serializers import Serializer
from wikidataintegrator.wdi_core import WDItemEngine

from .fields import (
    WikidataAltLabelField,
    WikidataConformanceField,
    WikidataEntityField,
    WikidataField,
    WikidataLabelField
)
from .utils import (
    dict_has_substring,
    set_kwargs
)
from .viewsets import generate_wikidata_item_viewset

_logger = logging.getLogger(__name__)


# TODO: Add Language Support
# TODO: Support nested model responses
# TODO: (currently child must declare a "main" in order to build query in proper order)
class WikidataItemBase(object):
    """ Base Wikidata Item Model """
    model_name = 'Wikidata Item'
    model_name_plural = 'Wikidata Items'
    main = WikidataEntityField(triples=[], required=True)
    label = WikidataLabelField(required=True)
    alt_labels = WikidataAltLabelField()
    schema = None
    # Instance Attributes set dynamically:
    id = None
    conformance = WikidataConformanceField()

    def __init__(self, **kwargs):
        for key, field in self.get_fields(with_keys=True):
            field.name = key
            setattr(self, key, None)
        set_kwargs(self, kwargs)

    @classmethod
    def get_fields(cls, with_keys=False):
        """
        Get all fields associated with this model.
        Args:
            with_keys (Optional[bool]): True to make the return zipped tuples by keys,
                                        False if iterable fields

        Returns (Union[zip[str, WikidataField], List[WikidataField]]):

        """
        fields = []
        keys = []
        for sub_cls in cls.__mro__:
            for key, field in sub_cls.__dict__.items():
                if isinstance(field, WikidataField) and key not in keys:
                    keys.append(key)
                    fields.append(field)
        return zip(keys, fields) if with_keys else fields

    @classmethod
    def build_serializer(cls):
        """
        Build a serializer class to support this object.
        Returns (Class[Serializer]):

        """
        attrs = {}
        for key, field in cls.get_fields(with_keys=True):
            # TODO: Addd documentation explaining relationship between main and id
            if key == 'main':
                key = 'id'
            attrs[key] = field.serializer
        return type('{}Serializer'.format(cls.__name__), (Serializer,), attrs)

    @classmethod
    def get_all(cls, with_conformance=False, limit=None):
        """
        Query Wikidata to get all instances of this model.
        Notes:
            TODO: Consider structuring as a generator return like Django does, and use the .objects.all() style
        Args:
            with_conformance (Optional[Bool]): True if intending to use SheX validation, False otherwise
            limit (Optional[Int]): The maximum records to query Wikidata for

        Returns (List[WikidataItemBase]): Returns a list of this model's instances

        """
        wikidata_response = cls._query_wikidata(limit=limit)
        return [cls._from_wikidata(x, with_conformance) for x in wikidata_response]

    @classmethod
    def get(cls, entity_id, with_conformance=False):
        """
        Query Wikidata to get an instance by Wikidata Entity id.
        Notes:
            TODO: Consider use the .objects.get() style
        Args:
            entity_id (str): Wikidata Entity identifier (ex. "Q1234")
            with_conformance (Optional[Bool]): True if intending to use SheX validation, False otherwise

        Returns (Optional[WikidataItemBase]): Returns this model's instance with id=entity_id if found, None otherwise

        """

        wikidata_response = cls._query_wikidata((entity_id,), limit=1)
        if wikidata_response:
            return cls._from_wikidata(wikidata_response[0], with_conformance)
        _logger.warning("Unable to find %s with Wikidata Entity ID '%s'", cls.model_name, entity_id)
        return None

    @classmethod
    def search(cls, search_string):
        """
        Get all instances of this model that have an attribute with the given search_string.
        Args:
            search_string (str): Used to perform a substring match of all objects

        Returns (List[WikidataItemBase]):

        """
        return [obj for obj in cls.get_all() if obj._has_substring(search_string)]

    @classmethod
    def build_query(cls, values=None, limit=None):
        """
        Build a SPARQL query to fetch data that instantiates instances of models from the Wikidata Query Service.
        Args:
            values (Optional[Iterable[str]]): list-like structure containing Wikidata Entity ID's, ex. ['Q123', 'Q321']
            limit (Optional[int]): used to set a maximum return value in the query

        Returns (str): minified SPARQL Query string

        """
        # TODO: Add Offset
        fields = cls().get_fields()
        to_fields = ' '.join(f.to_wikidata_field() for f in fields)
        to_filters = ' '.join(f.to_wikidata_filter() for f in fields)
        to_services = ' '.join(f.to_wikidata_service() for f in fields)
        to_group = ' '.join(f.to_wikidata_group() for f in fields)
        value_filter = "VALUES (?main) {{ (wd:{vals}) }}".format(vals=") (wd:".join(val for val in values)) \
            if values else ''
        limit_by = f"LIMIT {limit}" if limit else ""

        query = f"""
                SELECT DISTINCT {to_fields}
                WHERE {{
                    {value_filter}
                    {to_filters}
                    SERVICE wikibase:label {{ bd:serviceParam wikibase:language "[AUTO_LANGUAGE],en". {to_services} }}
                }}
                GROUP BY {to_group}
                {limit_by}

            """
        return " ".join(query.split())

    @classmethod
    def get_viewset_urls(cls, slug='wikidata_item'):
        """
        Get viewset router urls to append to the django URLconf.
        Notes:
            For reference on url return types, check out the django docs:
            https://docs.djangoproject.com/en/2.2/ref/urls/
        Args:
            slug (Optional[str]): Base slug to set the path of the url patterns.

        Returns (URLResolver): An object that can be added to urlpatterns

        """
        return generate_wikidata_item_viewset(cls, slug)

    @classmethod
    def _query_wikidata(cls, values=None, limit=None):
        """
        Query Wikidata for data related to this model.
        Args:
            values (Optional[Iterable[str]]): list-like structure containing Wikidata Entity ID's, ex. ['Q123', 'Q321']
            limit (Optional[int]): used to set a maximum return value in the query

        Returns (List[Dict]]): The results->bindings of the Wikidata Query Service response

        """
        # TODO: Add some custom error handling
        results_json = WDItemEngine.execute_sparql_query(cls.build_query(values, limit))
        return results_json['results']['bindings']

    @classmethod
    def _from_wikidata(cls, wikidata_response, with_conformance=False):
        """
        Parse Wikidata Query Service response to instantiate a model object.
        Args:
            wikidata_response (Dict[str, Dict[str, str]]):
            with_conformance (Optional[Bool]): True if intending to use SheX validation, False otherwise

        Returns (WikidataItemBase):

        """
        obj = cls()
        for field in cls.get_fields():
            setattr(obj, field.name, field.from_wikidata(wikidata_response))
        obj.id = obj.main
        assert obj.id, "Wikidata Item Must Have Identifier"
        return obj.set_conformance() if with_conformance else obj

    def _has_substring(self, substring):
        return dict_has_substring(self.__dict__, substring)

    def __repr__(self):
        return "<{}: {}>".format(self.model_name, self.__str__())

    def __str__(self):
        return "{} ({})".format(self.label, self.main)

    def set_conformance(self):
        if self.schema:
            self.conformance = WDItemEngine.check_shex_conformance(self.id, self.schema, output="all")
        else:
            _logger.debug("No schema associated with this model: Setting result to 'n/a' "
                          "(Consider adding a model schema before using this functionality)")
            self.conformance = {
                'focus': self.id,
                'reason': 'No Schema associated with this model',
                'result': 'n/a'
            }
        return self


class WDTriple(object):
    def __init__(self, prop, values, subclass=False, minus=False):
        assert not (len(values) > 1 and minus), "Union and Minus should not be used in the same clause"
        self.prop = "{}/wdt:P279*".format(prop) if subclass else prop
        self.values = values
        query = " UNION ".join(f"{{{{ ?{{name}} wdt:{self.prop} wd:{val}.}}}}" for val in self.values)
        self._query = f"MINUS {query}" if minus else query

    def format(self, field_name):
        return self._query.format(name=field_name)



from django.conf.urls import url
from django.urls import include
from .views import schema_view

urlpatterns = [
    url(r'^$', schema_view.with_ui('redoc', cache_timeout=0), name='schema-redoc'),
    url(r'^swagger(?P<format>\.json|\.yaml)$', schema_view.without_ui(cache_timeout=0), name='schema-json'),
    url(r'^swagger/$', schema_view.with_ui('swagger', cache_timeout=0), name='schema-swagger-ui'),
    url(r'^api-auth/', include('rest_framework.urls'))
]



from rest_framework import serializers


class WikidataItemSerializer(serializers.Serializer):
    # TODO: Add QID validator and QIDField
    id = serializers.RegexField(regex="(Q|q)\d+", allow_blank=False, min_length=2, max_length=20,
                                help_text="Wikidata Item Identifier (ex. Q59961716)")
    label = serializers.CharField(allow_null=False, allow_blank=False)
    alt_labels = serializers.ListField(allow_null=True, allow_empty=True)

    def create(self, validated_data):
        # TODO: Add a create method that would call a .create method of the model
        pass

    def update(self, instance, validated_data):
        # TODO: Add a update method that would call a .update method of the model
        pass


class WikidataConformanceSerializer(serializers.Serializer):
    # TODO: Add QID validator and QIDField
    focus = serializers.RegexField(regex="(Q|q)\d+", allow_blank=False, min_length=2, max_length=20,
                                   help_text="Wikidata Item Identifier (ex. Q59961716)")
    reason = serializers.CharField(allow_null=False, allow_blank=False)
    result = serializers.BooleanField(allow_null=True)

    def create(self, validated_data):
        # TODO: Add a create method that would call a .create method of the model
        pass

    def update(self, instance, validated_data):
        # TODO: Add a update method that would call a .update method of the model
        pass



from rest_framework import serializers

from .serializers import WikidataConformanceSerializer
from .utils import (
    get_wikidata_field,
    set_kwargs
)

# TODO: Fields
# BindField - Example: BIND(wd:Q937 AS ?item).


class WikidataField(object):
    name = None
    serializer_class = serializers.Field
    default_serializer_settings = {}
    serializer = None

    def __init__(self, properties=None, values=None, default=None, required=False, entity_name='main',
                 serializer_settings=None, **kwargs):
        self.entity_name = entity_name
        self.properties = properties
        self.values = values
        self.default = default
        self.required = required
        set_kwargs(self, kwargs)
        self.set_serializer(serializer_settings or {})

    def __repr__(self):
        return "<{}: {}>".format(self.__class__.__name__, self.name)

    def set_serializer(self, serializer_settings):
        for key, value in self.default_serializer_settings.items():
            if key not in serializer_settings:
                serializer_settings[key] = value
        self.serializer = self.serializer_class(**serializer_settings)

    def _prop_sparql_string(self):
        return "wdt:{}".format('|wdt:'.join(self.properties))

    def to_wikidata_field(self):
        return "?{self.name}".format(self=self)

    def to_wikidata_filter(self):
        prop_string = self._prop_sparql_string()
        wd_triple = "?{self.entity_name} {props} ?{self.name}.".format(self=self, props=prop_string)
        return wd_triple if self.required else "OPTIONAL {{}}".format(wd_triple)

    def to_wikidata_service(self):
        return ""

    def to_wikidata_group(self):
        return ""

    def from_wikidata(self, wikidata_response):
        return get_wikidata_field(wikidata_response, self.name, self.default)


class WikidataListResponseMixin(object):
    separator = '|'
    name = None
    default = None
    serializer_class = serializers.ListField
    default_serializer_settings = {'allow_null': True, 'allow_empty': True}

    def from_wikidata(self, wikidata_response):
        field = get_wikidata_field(wikidata_response, self.name, self.default)
        return field.split(self.separator) if field else self.default


class WikidataStringField(WikidataField):
    pass


class WikidataLabelField(WikidataField):
    suffix = 'Label'
    serializer_class = serializers.CharField
    default_serializer_settings = {'allow_null': False, 'allow_blank': False}

    def __init__(self, **kwargs):
        super(WikidataLabelField, self).__init__(**kwargs)
        self.from_name = "{}{}".format(self.entity_name, self.suffix)

    def to_wikidata_field(self):
        return "?{self.from_name} (?{self.from_name} AS ?{self.name})".format(self=self)

    def to_wikidata_filter(self):
        return ''  # Labels are not in the WHERE clause in a SPARQL query

    def to_wikidata_service(self):
        # TODO: Merge similarities with entity list label
        return "?{self.entity_name} rdfs:label ?{self.from_name} . ".format(self=self)

    def to_wikidata_group(self):
        return "?{self.from_name}".format(self=self)


class WikidataEntityField(WikidataField):
    # TODO: Add Item and Property SubClasses
    serializer_class = serializers.RegexField
    default_serializer_settings = {'allow_blank': False, 'regex': "(Q|q)\d+", 'min_length': 2, 'max_length': 20,
                                   'help_text': "Wikidata Item Identifier (ex. Q59961716)"}
    wikidata_filter = None

    def __init__(self, triples, **kwargs):
        super(WikidataEntityField, self).__init__(**kwargs)
        self.wikidata_filter = " ".join(triple.format(self.entity_name) for triple in triples)

    def to_wikidata_filter(self):
        return self.wikidata_filter if self.required else "OPTIONAL {{ {} }}".format(self.wikidata_filter)

    def to_wikidata_group(self):
        return "?{self.name}".format(self=self)

    def from_wikidata(self, wikidata_response):
        field = super(WikidataEntityField, self).from_wikidata(wikidata_response)
        return field.replace('http://www.wikidata.org/entity/', '')


class WikidataEntityFilterField(WikidataField):
    def to_wikidata_filter(self):
        prop_string = self._prop_sparql_string()
        wd_filter = "?{self.entity_name} {props} ?{self.name}_qid. FILTER(?{self.name}_qid=wd:{vals}).".format(
            self=self, props=prop_string, vals="|| ?{self.name}_qid=wd:".join(self.values)).format(self=self)
        return wd_filter if self.required else "OPTIONAL {{ {} }}".format(wd_filter)

    def to_wikidata_service(self):
        # TODO: Merge similarities with entity list label
        return "?{self.name}_qid rdfs:label ?{self.name} . ".format(self=self)

    def to_wikidata_group(self):
        return "?{self.name}".format(self=self)


class WikidataListField(WikidataListResponseMixin, WikidataField):
    def __init__(self, properties=None, empty=True, separator='|', **kwargs):
        self.separator = separator
        if empty:
            kwargs['default'] = []
        super(WikidataListField, self).__init__(properties, **kwargs)

    def to_wikidata_field(self):
        return "(GROUP_CONCAT(DISTINCT ?{self.name}_item; SEPARATOR='{self.separator}') AS ?{self.name})".format(
            self=self)

    def to_wikidata_filter(self):
        prop_string = self._prop_sparql_string()
        wd_filter = "?{self.entity_name} {props} ?{self.name}_item .".format(self=self, props=prop_string)
        return wd_filter if self.required else "OPTIONAL {{ {} }}".format(wd_filter)


class WikidataAltLabelField(WikidataListField):
    suffix = '_alt_label'

    def to_wikidata_field(self):
        return "(GROUP_CONCAT(DISTINCT ?{self.entity_name}_alt_label; SEPARATOR='{self.separator}') AS " \
               "?{self.name})".format(self=self)

    def to_wikidata_filter(self):
        # TODO: add language support in here
        wd_filter = "?{self.entity_name} skos:altLabel ?{self.entity_name}_alt_label ." \
               "FILTER (lang(?{self.entity_name}_alt_label)='en')".format(self=self)
        return wd_filter if self.required else "OPTIONAL {{ {} }}".format(wd_filter)


class WikidataEntityListField(WikidataListField):
    def __init__(self, properties=None, **kwargs):
        assert properties, "There must be a list/tuple of properties"
        super(WikidataEntityListField, self).__init__(properties, **kwargs)

    def to_wikidata_field(self):
        return "(GROUP_CONCAT(DISTINCT ?{self.name}_itemLabel; SEPARATOR='{self.separator}') AS ?{self.name})".format(
            self=self)

    def to_wikidata_service(self):
        return "?{self.name}_item rdfs:label ?{self.name}_itemLabel . ".format(self=self)


class WikidataConformanceField(WikidataField):
    serializer_class = WikidataConformanceSerializer
    default_serializer_settings = {}

    # def __init__(self, **kwargs):
    #     super(WikidataLabelField, self).__init__(**kwargs)
    #     self.from_name = "{}{}".format(self.entity_name, self.suffix)

    def to_wikidata_field(self):
        return ""

    def to_wikidata_filter(self):
        return ""

    def to_wikidata_service(self):
        return ""

    def to_wikidata_group(self):
        return ""



def get_wikidata_field(wikidata_response_dict, key, default=None):
    # TODO: create a wikidata response object and move this to its instance method
    try:
        return wikidata_response_dict[key].get('value', default)
    except (KeyError, ValueError):
        return default


def get_wikidata_concat_list(wikidata_response_dict, key, default=None, separator='|'):
    if not default:
        default = []
    field = get_wikidata_field(wikidata_response_dict, key, default)
    return field.split(separator) if field else default


def dict_has_substring(haystack, needle):
    return all(token in str(haystack.values()).lower() for token in str(needle).lower().split())


def set_kwargs(obj, kwargs):
    for key, value in kwargs.items():
        setattr(obj, key, value)



# coding=utf-8
""" Setup script to build django-wikidata-api """
import os
from setuptools import setup

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.md')).read()

setup(
    name='django-wikidata-api',
    version='0.0.1',
    packages=['django_wikidata_api'],
    description='Python Package for interfacing with Wikidata within a Django App',
    long_description=README,
    long_description_content_type="text/markdown",
    author='Kenneth Seals-Nutt',
    author_email='kenneth@seals-nutt.com',
    url='https://github.com/kennethsn/django-wikidata-api',
    license='AGPL-3.0',
    install_requires=[
        'Django>=2.2,<2.3',
        'djangorestframework>=3.10.3',
        'drf-yasg>=1.17',
        'mock',
        'wikidataintegrator>=0.4.2, <0.4.3',
        # Wikidataintegrator packages that aren't tied down
        'pandas>=0.25.2',
        'tqdm>=4.36.1',
    ],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Natural Language :: English",
        "Operating System :: MacOS :: MacOS X",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Framework :: Django",
        "Framework :: Django :: 2.2",
        "License :: OSI Approved :: GNU Affero General Public License v3"
    ],
    python_requires='>=3.6',
)
